#                   Copyright (c) 2021, Serum Studio

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import os
from typing import Optional
from typing import TextIO

if os.name == "nt":
    import ctypes

    class _CursorInfo(ctypes.Structure):
        _fields_ = [("size", ctypes.c_int), ("visible", ctypes.c_byte)]


def hide(stream: Optional[TextIO] = sys.stdout):
    """
    Hide the cursor.
    Parameters
    ----------
    stream: (sys.stdout)
        Defines stream to write output to.

    """
    if os.name == "nt":
        ci = _CursorInfo()
        handle = ctypes.windll.kernel32.GetStdHandle(-11)
        ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
        ci.visible = False
        ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))

    elif os.name == "posix":
        stream.write("\033[?25l")
        stream.flush()


def show(stream: Optional[TextIO] = sys.stdout):
    """
    Show cursor.

    Parameters
    ----------
    stream: (sys.stdout)
        Defines stream to write output to.

    """
    if os.name == "nt":
        ci = _CursorInfo()
        handle = ctypes.windll.kernel32.GetStdHandle(-11)
        ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
        ci.visible = True
        ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))

    elif os.name == "posix":
        stream.write("\033[?25h")
        stream.flush()



from hype.errors import PluginError
from .style import Background, Color
from .style import Style

try:
    import hype.color

    COLOR_SUPPORTED = True
except PluginError:
    COLOR_SUPPORTED = False

rule_colors = {
    "red": Color.RED,
    "blue": Color.BLUE,
    "green": Color.GREEN,
    "black": Color.BLACK,
    "cyan": Color.CYAN,
    "magenta": Color.MAGENTA,
    "yellow": Color.YELLOW,
    "white": Color.WHITE,
    "reset": Color.RESET,
}

bg_colors = {
    "red": Background.RED,
    "blue": Background.BLUE,
    "green": Background.GREEN,
    "black": Background.BLACK,
    "cyan": Background.CYAN,
    "magenta": Background.MAGENTA,
    "yellow": Background.YELLOW,
    "white": Background.WHITE,
    "reset": Background.RESET,
}

rule_styles = {
    "bold": Style.BOLD,
    "italic": Style.ITALIC,
    "underline": Style.UNDERLINE,
    # ALiases
    "b": Style.BOLD,
    "i": Style.ITALIC,
    "u": Style.UNDERLINE,
}

all_tags = [
    #: colors
    "red",
    "blue",
    "green",
    "black",
    "cyan",
    "magenta",
    "yellow",
    "white",
    "reset",
    #: background
    "bg",
    #: styles
    "i",
    "italic",
    "b",
    "bold",
    "u",
    "underline",
]



import optparse
from os import name
from typing import Any, Dict
from typing import Optional
from typing import List
from typing import Tuple
import textwrap



class HypeArgument:
    """
    Base class for handling arguments.
    """
    __metavar_mapping = {
        str: "STRING",
        int: "INTEGER",
        float: "FLOAT",
        bool: "BOOLEAN",
        bytes: "BYTES",
        list: "LIST",
        dict: "DICTIONARY",
    }
    def __init__(
        self,
        name: str,
        help: str = None,
        type: Any = None
    ):

        self.name = name
        self.help = help
        self.type = self.__metavar_mapping[type] if type else None
        

class HypeOptionParser(optparse.OptionParser):
    """
    This is a parser inherited by `optparse.OptionParser`
    with some minor changes and adding arguments as well as 
    including help formatter.
    """
    __args = []
    def __init__(self, arguments: List[Any], *args, **options):
        super(HypeOptionParser, self).__init__(*args, **options)
        
        self.options = options
        self.arguments = arguments
        
        if self.arguments:
            self.usage = "%prog [ARGS] [OPTIONS]"
        else:
            self.usage = "%prog [OPTIONS]"

        self.disable_interspersed_args()

    def add_argument(self, argument):
        if not isinstance(argument, HypeArgument):
            raise ValueError('{} is not a instance of HypeArgument'.format(argument))
        
        return self.arguments.append(argument)

    def format_help(self, formatter=None):
        output = optparse.OptionParser.format_help(self, formatter)
        result = []
        display_names = []
        help_position = 0

        if formatter == None:
            formatter = self.formatter
        
        if self.arguments:
            result += ['\n']
            result += formatter.format_heading('Arguments')
            formatter.indent()

            for arg in self.arguments:
                name = arg.name
                if arg.type:
                    name = "{}={}".format(name, arg.type)
                display_names.append(name)

                #: Set the help position based on the max width.
                proposed_help_position = len(name) + formatter.current_indent + 2
                if proposed_help_position <= formatter.max_help_position:
                    help_position = max(help_position, proposed_help_position)  
                    
            
            for arg, name in zip(self.arguments, display_names):
                name_width = help_position - formatter.current_indent - 2

                if len(name) > name_width:
                    name = "%*s%s\n" % (formatter.current_indent, "", name)
                    indent_first = help_position

                else:
                    name = "%*s%-*s  " % (formatter.current_indent, "",
                                        name_width, name)
                    indent_first = 0

                result.append(name)
                help_width = formatter.width - help_position
                help_lines = textwrap.wrap(arg.help, help_width)
                result.append("%*s%s\n" % (indent_first, "", help_lines[0]))
                result.extend(["%*s%s\n" % (help_position, "", line)
                            for line in help_lines[1:]])
            
            
            formatter.dedent()

        result += ['\n']
        return output + ''.join(result)


class HypeCommand:
    """
    A command for Hype CLI. Commands are different with options.
    It act as a sub command argument that has different type of options


    Parameters:

        name (str):
            The name of the command

        usage (str):
            The usage format for the command

        help (str):
            The help for the command.


        aliases (tuple):
            Aliases of the command.


    Example:

        >>> greet = HypeCommand('greet', usage='%prog [OPTIONS]',
        ...      help="Greet the user", aliases=('g'))

        >>> greet.add_option('--name', type=str, required=True)

    """

    def __init__(
        self,
        name: str,
        usage: Optional[str] = None,
        aliases: Optional[Tuple[Any]] = None,
        help: Optional[str] = None,
        args: Optional[List[Any]] = None,
    ):
        self.name = name
        self.usage = usage
        self.aliases = aliases
        self.help = help or "This command accept a positional arguments"
        self.args = args
        self.parser = HypeOptionParser(self.args)
        

        if self.usage:
            self.parser.usage = self.usage

    def add_option(self, *args, **kwargs):
        return self.parser.add_option(*args, **kwargs)



#                   Copyright (c) 2021, Serum Studio

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#: The parser was inspired at pyskcode. Thanks to https://github.com/TamiaLab/PySkCode

import string
from .constants import rule_colors
from .constants import rule_styles
from .constants import bg_colors
from .constants import all_tags
from typing import Optional
from typing import IO
from .errors import PluginError
from .errors import TagNotFound

__all__ = ["print_color", "parsed_color"]

try:

    import colorama

    colorama.init()
    COLOR_SUPPORTED = True

except ModuleNotFoundError:
    COLOR_SUPPORTED = False


# Character charsets
WHITESPACE_CHARSET = frozenset(string.whitespace)
IDENTIFIER_CHARSET = frozenset(string.ascii_letters + string.digits + "_*")

# Token types
TOKEN_DATA = 0
TOKEN_NEWLINE = 1
TOKEN_OPEN_TAG = 2
TOKEN_CLOSE_TAG = 3
TOKEN_SELF_CLOSE_TAG = 4


def skip_whitespaces(text: str, offset: int):
    """
    Skip any whitespaces.
    :param text: The input text.
    :param offset: The current offset in the input text.
    :return The new offset in the input text.
    """
    while text[offset] in WHITESPACE_CHARSET:
        offset += 1
    return offset


def get_identifier(text: str, offset: int):
    """
    Get the identifier string starting in the text at the given offset.
    :param text: The input text.
    :param offset: The current offset in the input text.
    :return The identifier normalized as lowercase and the new offset in the input text.
    """
    identifier = ""

    # Get any identifier char
    while text[offset] in IDENTIFIER_CHARSET:
        identifier += text[offset]

        # Process the next char
        offset += 1

    # Normalize the identifier and return new values
    return identifier.lower(), offset


def get_attribute_value(
    text: str, offset: int, opening_tag_ch: str, closing_tag_ch: str
):
    """
    Get the attribute value starting in the text at the given offset.
    :param text: The input text.
    :param offset: The current offset in the input text.
    :param opening_tag_ch: The opening tag char (must be one char long).
    :param closing_tag_ch: The closing tag char (must be one char long).
    :return The attribute value with trailing whitespaces removed and the new offset in the input text.
    """
    attribute_value = ""

    # Handle quoted and unquoted value
    if text[offset] == "'" or text[offset] == '"':
        quoting_ch = text[offset]

        # Process the next char
        offset += 1

        # Get attribute value
        while text[offset] != quoting_ch:

            # Handle escape sequences
            if text[offset] == "\\":

                # Process the next char
                offset += 1

                # Only the quoting char and the backslash char can be escaped
                if text[offset] != quoting_ch and text[offset] != "\\":
                    attribute_value += "\\"

            # Store the char
            attribute_value += text[offset]

            # Process the next char
            offset += 1

        # Skip last quoting sign
        offset += 1

    else:

        # Get raw attribute value
        ch = text[offset]
        while (
            ch != closing_tag_ch
            and ch != opening_tag_ch
            and ch not in WHITESPACE_CHARSET
        ):
            attribute_value += ch

            # Process the next char
            offset += 1
            ch = text[offset]

    # Check format
    if text[offset] != closing_tag_ch and text[offset] not in WHITESPACE_CHARSET:
        raise ValueError(
            "A whitespace is mandatory after the attribute "
            "value if not followed by the closing tag char"
        )

    # Strip the attribute value and return new values
    return attribute_value.strip(), offset


def parse_tag(
    text: str,
    start_offset: int,
    opening_tag_ch="[",
    closing_tag_ch="]",
    allow_tagvalue_attr=True,
    allow_self_closing_tags=True,
):
    """
    Parse the text starting at ``start_offset`` to extract a valid tag, if any.
    Return the tag name, the ``is_closing_tag`` and ``is_self_closing_tag`` flags,
    the tag attributes (as a dictionary) and the offset just after the end of the tag.

    If something goes wrong (malformed tag, unexpected end-of-file, etc.) an exception
    of type``IndexError`` (unexpected end-of-file) or ``ValueError`` (malformed tag) is raised.

    :param text: The input text.
    :param start_offset: The offset in the input text to start with.
    :param opening_tag_ch: The opening tag char (must be one char long, default '[').
    :param closing_tag_ch: The closing tag char (must be one char long, default ']').
    :param allow_tagvalue_attr: Set to ``True`` to allow the BBcode ``tagname=tagvalue`` syntax shortcut

    (default is ``True``).

    :param allow_self_closing_tags: Set to ``True`` to allow the self closing tags syntax (default is ``True``).

    :return A tuple ``(tag_name, is_closing_tag, is_self_closing_tag, tag_attrs, offset + 1)`` on success, or an
    exception on error (see possible exception in the docstring above).
    """
    assert text, "No text input given (mandatory)."
    assert start_offset >= 0, "Starting offset must be greater or equal to zero."
    assert start_offset < len(
        text
    ), "Starting offset must be lower than the size of the text."
    assert (
        len(opening_tag_ch) == 1
    ), "Opening tag character must be one char long exactly."
    assert (
        len(closing_tag_ch) == 1
    ), "Closing tag character must be one char long exactly."
    assert (
        text[start_offset] == opening_tag_ch
    ), "Unexpected call to parse_tag, no opening tag char at starting offset."

    # Init tag variables
    tag_attrs = {}
    is_closing_tag = is_self_closing_tag = False

    # Skip the opening char and whitespaces
    offset = skip_whitespaces(text, start_offset + 1)

    # Check for closing tag
    if text[offset] == "/":

        # Look like a closing tag for me
        is_closing_tag = True

        # Skip slash and whitespaces
        offset = skip_whitespaces(text, offset + 1)

    # Get the tag name
    tag_name, offset = get_identifier(text, offset)

    # Detect invalid tag early
    if not tag_name:
        raise ValueError("Invalid tag format: no tag name found")

    # Skip whitespaces
    offset = skip_whitespaces(text, offset)

    # Check for closing char if is_closing_tag is set (closing tags have no attribute)
    if is_closing_tag and text[offset] != closing_tag_ch:
        raise ValueError("Invalid tag format: closing tags cannot have attribute")

    # Check for tag value
    if text[offset] == "=":

        # Check for support
        if not allow_tagvalue_attr:
            raise ValueError(
                "Invalid tag format: tagname=tagvalue shortcut support disabled by caller."
            )

        # Skip equal sign and whitespaces
        offset = skip_whitespaces(text, offset + 1)

        # Get the tag value
        tag_value, offset = get_attribute_value(
            text, offset, opening_tag_ch, closing_tag_ch
        )

        # Store the tag value
        tag_attrs[tag_name] = tag_value

        # Skip whitespaces
        offset = skip_whitespaces(text, offset)

    # Named attributes handling
    while text[offset] != closing_tag_ch and text[offset] != "/":

        # Get the attribute name
        attr_name, offset = get_identifier(text, offset)

        # Detect erroneous attribute name
        if not attr_name:
            raise ValueError(
                "Invalid tag format: no attribute name found or invalid character found"
            )

        # Skip whitespaces
        offset = skip_whitespaces(text, offset)

        # Check for attr value
        if text[offset] == "=":

            # Skip equal sign and whitespaces
            offset = skip_whitespaces(text, offset + 1)

            # Get the attribute value
            attr_value, offset = get_attribute_value(
                text, offset, opening_tag_ch, closing_tag_ch
            )

            # Store the tag value
            tag_attrs[attr_name] = attr_value

            # Skip whitespaces
            offset = skip_whitespaces(text, offset)

        else:

            # Attribute without value
            tag_attrs[attr_name] = ""

    # Handle self closing tag
    if text[offset] == "/":

        # Check for support
        if not allow_self_closing_tags:
            raise ValueError(
                "Invalid tag format: self closing tags support disabled by caller."
            )

        # Look like a self closing tag for me
        is_self_closing_tag = True

        # Skip slash and whitespaces
        offset = skip_whitespaces(text, offset + 1)

    # Assert end of tag
    if text[offset] != closing_tag_ch:
        raise ValueError("Invalid tag format: malformed end of tag")

    # Emit the tag
    return tag_name, is_closing_tag, is_self_closing_tag, tag_attrs, offset + 1


def tokenize_newline(data: str):
    """
    Given a string that does not contain any tags, this function will
    yield a list of ``TOKEN_NEWLINE`` and ``TOKEN_DATA`` tokens in such way
    that if you concatenate their data, you will have the original string.
    N.B. Newline must have been normalized to ``\n`` before calling this function.
    :param data: Input data string to be tokenize.
    """
    lines = data.split("\n")
    last_line = lines.pop()
    for line in lines:
        if line:
            yield TOKEN_DATA, None, None, line
        yield TOKEN_NEWLINE, None, None, "\n"

    if last_line:
        yield TOKEN_DATA, None, None, last_line


def tokenize_tag(
    text: str,
    opening_tag_ch="[",
    closing_tag_ch="]",
    allow_tagvalue_attr=True,
    allow_self_closing_tags=True,
):
    """
    Split the given text into tokens (generator function).
    :param text: The input text to be tokenize.
    :param opening_tag_ch: The opening tag char (must be one char long, default '[').
    :param closing_tag_ch: The closing tag char (must be one char long, default ']').
    :param allow_tagvalue_attr: Set to ``True`` to allow the BBcode ``tagname=tagvalue`` syntax shortcut
    (default is ``True``).
    :param allow_self_closing_tags: Set to ``True`` to allow the self closing tags syntax (default is ``True``).
    """
    assert text, "No text input given (mandatory)."
    assert (
        len(opening_tag_ch) == 1
    ), "Opening tag character must be one char long exactly."
    assert (
        len(closing_tag_ch) == 1
    ), "Closing tag character must be one char long exactly."

    # Normalize newlines (fastest method)
    text = text.replace("\r\n", "\n").replace("\r", "\n")

    # Search an opening tag
    start = text.find(opening_tag_ch)
    pos = 0

    # Process the whole text until no more tag can be found
    # N.B. string.find(s, beg) return -1 if beg >= len(string)
    while start >= 0:

        # Run the parser to get the tag
        try:
            tag = parse_tag(
                text,
                start,
                opening_tag_ch,
                closing_tag_ch,
                allow_tagvalue_attr,
                allow_self_closing_tags,
            )

        except (IndexError, ValueError):

            # Continue searching if not a valid tag
            start = text.find(opening_tag_ch, start + 1)
            continue

        # Get the text before the tag, if any
        if start > pos:
            text_before_tag = text[pos:start]
            yield from tokenize_newline(text_before_tag)

        # Unpack the tag structure
        tag_name, is_closing_tag, is_self_closing_tag, tag_attrs, offset = tag
        tag_source = text[start:offset]

        # Yield the tag token
        if is_self_closing_tag:
            yield TOKEN_SELF_CLOSE_TAG, tag_name, tag_attrs, tag_source
        elif is_closing_tag:
            yield TOKEN_CLOSE_TAG, tag_name, tag_attrs, tag_source
        else:
            yield TOKEN_OPEN_TAG, tag_name, tag_attrs, tag_source

        # Store the current position in text for next loop
        pos = offset

        # Search the next tag if any
        start = text.find(opening_tag_ch, offset)

    # Yield the remaining piece of text if any
    remaining_text = text[pos:]
    if remaining_text:
        yield from tokenize_newline(remaining_text)


def parse_color(text: str = ""):
    """
    Parsed the text of string that contains color tag into colored texts

    Parameters:
    ---
        text(str):
            The string to be parsed.

    Example:
    ---

        >>> string = parse_color('[red]This is red[/red]')
        >>> ...
        >>> #: You can use the parse_color to standard
        >>> #: printing or storing variables.
        >>> ...
        >>> print(string)

    """

    parsed_text = []
    tokens = tokenize_tag(text)

    for token in tokens:
        token_type, tag_name, tag_attr, token_source = token

        if token_type == TOKEN_OPEN_TAG:
            if tag_name not in all_tags:
                raise TagNotFound(
                    '\n\n"%s" is not found. check avaialble tags on the documentation:\nhttps://hype.serum.studio\n\n'
                    % (tag_name)
                )

            if tag_name in rule_colors:
                parsed_text.append(rule_colors[tag_name])

            if tag_name in rule_styles:
                parsed_text.append(rule_styles[tag_name])

            if tag_name == "bg":

                if "color" in tag_attr:
                    parsed_text.append(bg_colors[tag_attr.get("color")])

                else:
                    raise AttributeError(
                        "when defining background, color attribute must define."
                    )

        if token_type == TOKEN_CLOSE_TAG:
            parsed_text.append(rule_colors["reset"])

        if token_type == TOKEN_DATA:
            parsed_text.append(token_source)

        if token_type == TOKEN_NEWLINE:
            parsed_text.append(token_source)

    return "".join(parsed_text)


def print_color(
    text: str = "",
    sep: Optional[str] = " ",
    end: Optional[str] = "\n",
    file: Optional[IO[str]] = None,
    flush: Optional[bool] = False,
):

    """
    Simillar to built-in function, `print` but it prints a colored text
    from `parsed_color`

    Parameters:
    ---
        Similar to `print`

    Example:
    ---

        >>> print_color('[green]This is green[/green]')
        >>> #: Output This is green -> The output is colored in green.

    """

    if COLOR_SUPPORTED == False:
        raise PluginError(
            """

        -----------------------------------

        Plugin not installed properly:

        `color`: In order to install the `color` plugin,
        you may run the commmand `pip install hypecli[color]`
        or read the documentation.

        https://hype.serum.studio/
        
        -----------------------------------
        """
        )

    parsed_text = parse_color(text)

    print(parsed_text, sep=sep, end=end, file=file, flush=flush)




#                   Copyright (c) 2021, Serum Studio

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import List
from hype.command import HypeCommand
import optparse
from optparse import HelpFormatter
import sys
import textwrap

class HypeParser(optparse.OptionParser):
    """
    A command parser for Hype CLI that was built on the top of 
    `optparse.OptionParser` This parser is pretty simmilar with 
    OptionParser, the only difference is the commands. 
    

    Parameters:

        commands (list):
            A list of all HypeCommands
        
        **options (dict):
            A dictionary of kwargs

    Example:

        >>> greet = HypeCommand('greet', help="%prog [OPTIONS]")
        >>> greet.add_option('--name', type=str)
        >>> ...
        >>> goodbye = HypeCommand('goodbye', help="%prog [OPTIONS]")
        >>> goodbye.add_option('--name', type=str)
        >>> ...
        >>> parser = HypeParser( commands=(greet, goodbye) )
        >>> options, commands, command_opt, args = parser.parse_args()

    """
    _HelpCommand = HypeCommand('help', help="All details about the commands", aliases=('?'))

    def __init__(
        self, 
        commands: List[HypeCommand] = [], 
        *args, 
        **options
    ):
        
        self.commands = commands
        self.commands.append(self._HelpCommand)
        self.options = options

        if 'usage' not in self.options:
            self.options['usage'] = "%prog COMMAND [ARGS..]\n%prog help COMMAND"

        super(HypeParser, self).__init__(*args, **options)

        for command in self.commands:
            command.parser.prog = "%s %s" % (self.get_prog_name(), command.name)

        self.disable_interspersed_args()

    def add_command(self, cmd: HypeCommand):
        """
        Add a command. 
        
        Parameters
        ---
            cmd (HypeCommand):
                The command to be add.

        Example:

            >>> goodbye = HypeCommand(..)
            >>> parser = HyperParser(...)
            >>> ...
            >>> parser.add_command(goodbye)

        """
        if not isinstance(cmd, HypeCommand):
            raise TypeError('{} is not a instance of HypeCommand'.format(cmd))

        self.commands.append(cmd)


    def format_help(self, formatter=None) -> str:
        out = optparse.OptionParser.format_help(self, formatter)
        
        if formatter == None:
            formatter = self.formatter

        #: HEADER for the Help command
        result = ['\n']
        result.append(formatter.format_heading('Commands'))
        formatter.indent()

        display_names = []
        help_position = 0

        for command in self.commands:
            name = command.name
            if command.aliases:
                #: Add aliases of the command
                name += ' (%s)' % (', '.join(command.aliases))

            display_names.append(name)

            #: Set the help position based on the max width.
            proposed_help_position = len(name) + formatter.current_indent + 2
            if proposed_help_position <= formatter.max_help_position:
                help_position = max(help_position, proposed_help_position)  
    
        #: Add the command to the output
        for command, name in zip(self.commands, display_names):
            #: From optparse.py

            name_width = help_position - formatter.current_indent - 2

            if len(name) > name_width:
                name = "%*s%s\n" % (formatter.current_indent, "", name)
                indent_first = help_position

            else:
                name = "%*s%-*s  " % (formatter.current_indent, "",
                                      name_width, name)
                indent_first = 0

            result.append(name)
            help_width = formatter.width - help_position
            help_lines = textwrap.wrap(command.help, help_width)
            result.append("%*s%s\n" % (indent_first, "", help_lines[0]))
            result.extend(["%*s%s\n" % (help_position, "", line)
                           for line in help_lines[1:]])

        result += ['\n']
        formatter.dedent()

        # Concatenate the original help message with the command list.
        return out + "".join(result)

    def __command_for_name(self, name):
        """
            Return the command in self.commands matching the
        given name. The name may either be the name of a subcommand or
        an alias. If no subcommand matches, returns None.
        
        Parameters:
            name (str):
                The name of the command to be matched.

        """ 

        _command = None

        for command in self.commands: 
            try:
            
                if name == command.name or name in command.aliases:
                    _command = command
            
            except TypeError:
                pass

        return _command

    def parse_args(self, _args=None, _value=None):
        """
        Just like the `parse_args` from OptionParser but add some more value.
        
        Added Value:
        ---

        - options: The option passed to the root parser
        - command: the command object that was invoked
        - command_opt: The option parsed to the command parser
        - command_args: The positional arguments passed to the sub command
        
        Parameters:
        ---

            _args (any):
                inherited from `optparse.OptionParser.parse_args`

            _value (any):
                inherited from `optparse.OptionParser.parse_args`

        Example:
        ---

            >>> parser = HypeParser(...)
            >>> parser.add_option(...)
            >>> ...
            >>> options, command, \ 
            ...    command_opt, command_args = parser.parse_args()
    
        """ 
        options, args = optparse.OptionParser.parse_args(self, _args, _value)

        if not args:
            # No command given, show the help message
            self.print_help()
            self.exit()

        else:
            command_name = args.pop(0)
            command = self.__command_for_name(command_name)

            if not command:
                self.error('Unknown Command: {}'.format(command_name))

        command_opt, command_args = command.parser.parse_args(args)

        if command is self._HelpCommand:
            if command_args:

                command_name = command_args[0]

                #: Check for the help command on the command arguments.
                helpcommand = self.__command_for_name(command_name)
                helpcommand.parser.print_help()
                self.exit()

            else:
                self.print_help()
                self.exit()

        
        return options, command, command_opt, command_args



#                   Copyright (c) 2021, Serum Studio

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from hype.errors import *

import math
import sys
import threading
import time
import warnings
from contextlib import contextmanager
from itertools import chain
from itertools import islice
from itertools import repeat

try:
    from alive_progress.core.configuration import config_handler
    from alive_progress.animations.utils import spinner_player

    from alive_progress.core.logging_hook import install_logging_hook
    from alive_progress.core.logging_hook import uninstall_logging_hook

    from alive_progress.core.timing import gen_simple_exponential_smoothing_eta
    from alive_progress.core.timing import to_elapsed_text
    from alive_progress.core.timing import to_eta_text

    from alive_progress.core.utils import clear_traces
    from alive_progress.core.utils import render_title
    from alive_progress.core.utils import sanitize_text_marking_wide_chars
    from alive_progress.core.utils import get_terminal_size

except ModuleNotFoundError:
    pass

from hype.cursor import hide as hide_cursor
from hype.cursor import show as show_cursor
from hype.constants import rule_colors
from hype.constants import COLOR_SUPPORTED
from hype.errors import PluginError


@contextmanager
def progressbar(
    total=None, title=None, calibrate=None, bar_color=None, title_color=None, **options
):
    """
    Hype Progressbar inherited from `alive_bar` and add some Features.

    Parameters:
    ---
        Similar to `alive_bar` but then added 2 options:

        - `bar_color (str)`:
            Set the color of the bar.

        - `title_color (str)`:
            Set the title color.


    An alive progress bar to keep track of lengthy operations.
    It has a spinner indicator, elapsed time, throughput and ETA.
    When the operation finishes, a receipt is displayed with statistics.
    If the code is executed in a headless environment, ie without a
    connected tty, all features are disabled but the final receipt.
    Another cool feature is that it tracks the actual count in regard of the
    expected count. So it will look different if you send more (or less) than
    expected.

    Also, the bar installs a hook in the system print function that cleans
    any garbage out of the terminal, allowing you to print() effortlessly
    while using the bar.
    Use it like this:

    >>> from alive_progress import alive_bar
    ... with alive_bar(123, 'Title') as bar:  # <-- expected total and bar title
    ...     for item in <iterable>:
    ...         # process item
    ...         bar()  # makes the bar go forward
    The `bar()` method should be called whenever you want the bar to go forward.
    You usually call it in every iteration, but you could do it only when some
    criteria match, depending on what you want to monitor.
    While in a progress bar context, you have two ways to output messages:
      - the usual Python `print()` statement, which will properly clean the line,
        print an enriched message (including the current bar position) and
        continue the bar right below it;
      - the `bar.text('message')` call, which sets a situational message right within
        the bar, usually to display something about the items being processed or the
        phase the processing is in.
    If the bar is over or underused, it will warn you!
    To test all supported scenarios, you can do this:
    >>> for x in 1000, 1500, 700, 0:
    ...    with alive_bar(x) as bar:
    ...        for i in range(1000):
    ...            time.sleep(.005)
    ...            bar()
    Expected results are these (but you have to see them in motion!):
    |████████████████████████████████████████| 1000/1000 [100%] in 6.0s (167.93/s)
    |██████████████████████████▋⚠            | (!) 1000/1500 [67%] in 6.0s (167.57/s)
    |████████████████████████████████████████✗ (!) 1000/700 [143%] in 6.0s (167.96/s)
    |████████████████████████████████████████| 1000 in 5.8s (171.91/s)
    Args:
        total (Optional[int]): the total expected count

        title (Optional[str]): the title, will be printed whenever there's no custom message

        calibrate (int): maximum theoretical throughput to calibrate animation speed
            (cannot be in the global configuration because it depends on the current mode)

        **options: custom configuration options, which override the global configuration:

            length (int): number of characters to render the animated progress bar

            spinner (Union[str, object]): the spinner to be used in all renditions
                it's a predefined name in `show_spinners()`, or a custom spinner

            bar (Union[str, object]): bar to be used in definite and both manual modes
                it's a predefined name in `show_bars()`, or a custom bar

            unknown (Union[str, object]): bar to be used in unknown mode (whole bar is a spinner)
                it's a predefined name in `show_spinners()`, or a custom spinner

            theme (str): theme name in alive_progress.THEMES

            force_tty (bool): runs animations even without a tty (pycharm terminal for example)

            manual (bool): set to manually control percentage

            enrich_print (bool): includes the bar position in print() and logging messages

            title_length (int): fixed title length, or 0 for unlimited
    """

    if bar_color or title_color and not COLOR_SUPPORTED:
        raise PluginError(
            "You need color plugin inorder to style the bar. Read the docs for more info"
        )

    if total is not None:
        if not isinstance(total, int):
            raise TypeError(
                "integer argument expected, got '{}'.".format(type(total).__name__)
            )
        if total <= 0:
            total = None

    try:
        config = config_handler(**options)
    except NameError:
        raise PluginError(
            """

        -----------------------------------

        Plugin not installed properly:

        `progress`: In order to install the `progress` plugin,
        you may run the commmand `pip install hypecli[progress]`
        or read the documentation.

        https://hype.serum.studio/
        
        -----------------------------------
        """
        )

    def run(spinner):
        player = spinner_player(spinner)
        while thread:
            release_thread.wait()
            alive_repr(next(player))
            time.sleep(1.0 / fps())

    def alive_repr(spin=""):

        elapsed = time.time() - run.init
        run.rate = current() / elapsed if elapsed else 0.0

        if bar_color != None or title_color != None:
            #: Set the color for the progress bar.

            line = " ".join(
                filter(
                    None,
                    (
                        f"{rule_colors[title_color]}{title}{rule_colors['reset']}",
                        f"{rule_colors[bar_color]}{bar_repr(run.percent, end)}{rule_colors['reset']}",
                        spin,
                        monitor(),
                        "in",
                        to_elapsed_text(elapsed, end),
                        stats(),
                        run.text,
                    ),
                )
            )

        else:
            line = " ".join(
                filter(
                    None,
                    (
                        title,
                        bar_repr(run.percent, end),
                        spin,
                        monitor(),
                        "in",
                        to_elapsed_text(elapsed, end),
                        stats(),
                        run.text,
                    ),
                )
            )

        line_len, (cols, _) = len(line), get_terminal_size()
        with print_lock:
            if line_len < run.last_line_len:
                clear_traces()

            sys.__stdout__.write(line[:cols] + (spin and "\r" or "\n"))
            sys.__stdout__.flush()

        run.last_line_len = line_len

    def flush_buffer():
        if print_buffer:
            print()

    def set_text(message):
        run.text = sanitize_text_marking_wide_chars(message)

    if config.manual:
        # FIXME update bar signatures and remove deprecated in v2.
        def bar_handle(perc=None, text=None):
            """Bar handle for manual (bounded and unbounded) modes.
            Only absolute positioning.
            """
            if perc is not None:
                flush_buffer()
                run.percent = max(0.0, float(perc))  # ignores negative numbers.
            else:
                warnings.warn(
                    DeprecationWarning(
                        "percent will be mandatory in manual bar(),"
                        " please update your code."
                    ),
                    stacklevel=2,
                )
            update_hook()
            if text is not None:
                warnings.warn(
                    DeprecationWarning(
                        "use bar.text('') instead of bar(text=''),"
                        " please update your code."
                    ),
                    stacklevel=2,
                )
                set_text(text)

    else:

        def bar_handle(text=None, incr=1):
            """Bar handle for definite and unknown modes.
            Only relative positioning.
            """
            flush_buffer()
            # FIXME it was accepting 0 before, so a user could be using that to change text only
            run.count += max(0, int(incr))  # ignores negative numbers.
            update_hook()
            if text is not None:
                warnings.warn(
                    DeprecationWarning(
                        "use bar.text('') instead of bar(text=''),"
                        " please update your code."
                    ),
                    stacklevel=2,
                )
                set_text(text)

    def print_hook(part):
        if part != "\n":
            # this will generate a sequence of lines interspersed with None, which will later
            # be rendered as the indent filler to align additional lines under the same header.
            gen = chain.from_iterable(zip(repeat(None), part.splitlines(True)))
            print_buffer.extend(islice(gen, 1, None))
        else:
            header = header_template.format(run.count)
            nested = "".join(line or " " * len(header) for line in print_buffer)
            with print_lock:
                clear_traces()
                sys.__stdout__.write("{}{}\n".format(header, nested))
            print_buffer[:] = []

    print_buffer, print_lock = [], threading.Lock()
    header_template = "on {}: " if config.enrich_print else ""
    print_hook.write = print_hook
    print_hook.flush = lambda: None
    print_hook.isatty = sys.__stdout__.isatty

    def start_monitoring(offset=0.0):
        hide_cursor()
        sys.stdout = print_hook
        run.before_handlers = install_logging_hook()
        release_thread.set()
        run.init = time.time() - offset

    def stop_monitoring():
        show_cursor()
        sys.stdout = sys.__stdout__
        uninstall_logging_hook(run.before_handlers)  # noqa
        return time.time() - run.init

    thread, release_thread = None, threading.Event()
    if sys.stdout.isatty() or config.force_tty:

        @contextmanager
        def pause_monitoring():
            release_thread.clear()
            offset = stop_monitoring()
            alive_repr()
            yield
            start_monitoring(offset)

        bar_handle.pause = pause_monitoring
        thread = threading.Thread(target=run, args=(config.spinner(),))
        thread.daemon = True
        thread.start()

    if total or not config.manual:  # we can count items.
        logic_total, rate_spec, factor, current = (
            total,
            "f",
            1.0e6,
            lambda: run.count,
        )  # noqa
    else:  # there's only a manual percentage.
        logic_total, rate_spec, factor, current = (
            1.0,
            "%",
            1.0,
            lambda: run.percent,
        )  # noqa

    bar_handle.text, bar_handle.current = set_text, current
    if total or config.manual:  # we can track progress and therefore eta.
        spec = "({{:.1{}}}/s, eta: {{}})".format(rate_spec)
        gen_eta = gen_simple_exponential_smoothing_eta(0.5, logic_total)
        gen_eta.send(None)
        stats = lambda: spec.format(
            run.rate, to_eta_text(gen_eta.send((current(), run.rate)))
        )
        bar_repr = config.bar(config.length)
    else:  # unknown progress.
        bar_repr = config.unknown(config.length, config.bar)
        stats = lambda: "({:.1f}/s)".format(run.rate)  # noqa
    stats_end = lambda: "({:.2{}}/s)".format(run.rate, rate_spec)  # noqa

    # calibration of the dynamic fps engine.
    # I've started with the equation y = log10(x + m) * k + n, where:
    #   y is the desired fps, m and n are horizontal and vertical translation,
    #   k is a calibration factor, computed from some user input c (see readme for details).
    # considering minfps and maxfps as given constants, I came to:
    #   fps = log10(x + 1) * k + minfps, which must be equal to maxfps for x = c,
    # so the factor k = (maxfps - minfps) / log10(c + 1), and
    #   fps = log10(x + 1) * (maxfps - minfps) / log10(c + 1) + minfps
    # neat! ;)

    min_fps, max_fps = 2.0, 60.0
    calibrate = max(0.0, calibrate or factor)
    adjust_log_curve = 100.0 / min(calibrate, 100.0)  # adjust curve for small numbers
    factor = (max_fps - min_fps) / math.log10((calibrate * adjust_log_curve) + 1.0)

    def fps():
        if run.rate <= 0:
            return 10.0  # bootstrap speed
        if run.rate < calibrate:
            return math.log10((run.rate * adjust_log_curve) + 1.0) * factor + min_fps
        return max_fps

    end, run.text, run.last_line_len = False, "", 0
    run.count, run.percent, run.rate, run.init = 0, 0.0, 0.0, 0.0

    if total:
        if config.manual:

            def update_hook():
                run.count = int(math.ceil(run.percent * total))

        else:

            def update_hook():
                run.percent = run.count / total

        monitor = lambda: "{}{}/{} [{:.0%}]".format(  # noqa
            "(!) " if end and run.count != total else "", run.count, total, run.percent
        )
    elif config.manual:
        update_hook = lambda: None  # noqa
        monitor = lambda: "{}{:.0%}".format(  # noqa
            "(!) " if end and run.percent != 1.0 else "", run.percent
        )
    else:
        run.percent = 1.0
        update_hook = lambda: None  # noqa
        monitor = lambda: "{}".format(run.count)  # noqa

    title = render_title(title, config.title_length)
    start_monitoring()
    try:
        yield bar_handle
    finally:
        flush_buffer()
        stop_monitoring()
        if thread:
            local_copy = thread
            thread = None  # lets the internal thread terminate gracefully.
            local_copy.join()

        end, run.text, stats = True, "", stats_end
        alive_repr()



#                   Copyright (c) 2021, Serum Studio

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import Optional
from typing import Any
from typing import Callable
import inspect
import typing


def convert_param_to_option(param: str = None) -> str:
    if len(param) > 1:
        fmt_str = "--%s" % (param)

    else:
        fmt_str = "-%s" % (param)

    return fmt_str


def convert_option_to_string(option: str = None) -> str:
    return option.split("--")[1]


def create_bool_option(option: str = None) -> str:
    """
    Create --formal / --no-formal
    #: --no-formal is currently not avaialable. Just incase on future.
    """
    if option.startswith("--"):
        option = convert_option_to_string(option)

    else:
        option = option

    return ("--%s" % (option), "--no-%s" % (option))


class ParamOption:

    __metavar_mapping = {
        str: "STRING",
        int: "INTEGER",
        float: "FLOAT",
        bool: "BOOLEAN",
        bytes: "BYTES",
        list: "LIST",
        dict: "DICTIONARY",
    }

    def __init__(
        self,
        name: str = None,
        required: bool = None,
        default: Any = None,
        _type: type = None,
        dest: str = None,
        action: str = None,
    ):

        self.name = name
        self.required = required
        self.default = default
        self.type = _type
        self.action = action
        self.dest = dest
        self.metavar = self.__metavar_mapping[_type] if _type else None

        if self.type == bool:

            self.type = None

            if self.default == True:
                self.action = "store_false"

            else:
                self.action = "store_true"

    @property
    def to_dict(self):
        return {
            "name": self.name,
            "dest": self.dest,
            "metavar": self.metavar,
            "required": self.required,
            "default": self.default,
            "type": self.type,
            "action": self.action,
        }


class OptionDict:
    def __init__(
        self,
        name: str = None,
        default: Any = None,
        required: bool = None,
        type: type = None,
    ):

        self.name = name
        self.default = default
        self.type = type
        self.required = required

    @property
    def to_dict(self):
        return {
            "name": self.name,
            "default": self.default,
            "required": self.required,
            "type": self.type,
        }


class CommandDict:
    def __init__(
        self,
        name: str,
        usage: str = None,
        help: str = None,
        aliases: tuple = None,
        opt: list = [],
        func: Callable[..., Any] = None,
    ):

        self.name = name
        self.usage = usage
        self.help = help
        self.aliases = aliases
        self.opt = opt
        self.func = func

    @property
    def to_dict(self):
        return {
            "name": self.name,
            "usage": self.usage,
            "help": self.help,
            "aliases": self.aliases,
            "options": self.opt,
            "func": self.func,
        }



#                   Copyright (c) 2021, Serum Studio

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#: This source file is responsibble for printing output with styles and colors.

from typing import IO
from typing import Any
from typing import Optional
from builtins import print as _print
from .errors import PluginError
import sys

try:
    from .color import print_color
except PluginError:
    print_color = None


def print(
    value: str,
    sep: Optional[str] = " ",
    end: Optional[str] = "\n",
    file: Optional[IO[str]] = None,
    flush: Optional[bool] = False,
):

    """
    A wrapper for both color printing from `hype.color.print_color`
    and a standart print function.

    Parameters:
    ---
        Same as print().

    Example:
    ---

        >>> from hype import print
        >>> print('[red]This is red[/]') # Hype Color should be supported.
        >>> print('No Color, standart print function') # No color installed.

    """

    try:

        print_color(value, sep=sep, end=end, file=file, flush=flush)

    except Exception:

        _print(value, sep=sep, end=end, file=file, flush=flush)




#                   Copyright (c) 2021, Serum Studio

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from setuptools import setup, find_packages
from hype import __license__, __author__, __version__, __desc__

BASE_URL = "https://github.com/serumstudio/hype"

def get_long_description():

    with open("README.md", encoding="utf-8") as f:
        readme = f.read()

    return readme

extras_require = {

    'color': ['colorama==0.4.4'], #: Color support
    'standard': ['colorama==0.4.4'], #: Standard installation with color support
    'progress': ['alive-progress==1.6.2'], #: With progressbar support
    'table': ['tabulate==0.8.9'] #: With Table support
}



setup(
    name = "hypecli",
    author = __author__,
    description =__desc__,
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    project_urls={
        'Documentation': 'https://hype.serum.studio',
        'Source': BASE_URL,
        'Tracker': "%s/issues" % (BASE_URL)
    },
    version = __version__,
    license = __license__,
    url=BASE_URL,
    keywords='cli,commandline-toolkit,command line toolkit,python cli,python 3'.split(','),
    packages = [p for p in find_packages() if 'hype.prompt' not in p and 'hype.prompt.getkey' not in p],
    extras_require = extras_require,
    classifiers = [
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development",
        "Typing :: Typed",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License"
    ],
)


#                   Copyright (c) 2021, Serum Studio

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from hype.errors import PluginError
from typing import Any, Optional
from typing import List

try:
    from hype.color import parse_color
except PluginError:
    parse_color = None
try:
    from tabulate import tabulate

except ModuleNotFoundError:
    tabulate = None


class Table:
    """
    A table wrapper for `tabulate`

    Parameters:
    ---

        headers (List[str]):
            A list of header string. You can set it as None since there
            is a function `add_header` for adding header.

        rows (List[str]):
            A list of row to be added. You can set it as None since there
            is a function called `add_row` for adding row individually.

        type (str):
            A table type to be rendered. See the docs for more information.

    Functions:
    ---

        add_header (List[str]):
            Add a list of header (str) to the data.

        add_row (List[str]):
            Add a list of row (str) to the data.



    Example:
    ---
        >>> table = Table(headers=['Name', 'Age', 'Hobby'])
        >>> table.add_row(['Zenqi', '5', 'Programming'])
        >>> print(table.render()) #: or print(table())

    """

    #: Initialize empty header list
    __headers = []

    #: Initialize empty row list
    __rows = []

    #: Initialize empty type
    __type = None

    def __init__(
        self,
        headers: Optional[List[str]] = None,
        rows: Optional[List[str]] = None,
        type: Optional[str] = "fancy_grid",
    ):

        self.__headers = headers
        self.__rows = rows or []
        self.__type = type

    def add_row(self, row: Any):
        """
        Add a list row to the table data.

        Parameters:
        ---

            row (Any):
                Add a row to the data. It can be a list or a single string.

        """

        if parse_color:
            try:
                for i in range(len(row)):
                    row[i] = parse_color(row[i])
            except Exception:
                row = row

        self.__rows.append(row)

    def add_header(self, header: Any):
        """
        Add a list row to the table data.

        Parameters:
        ---

            header (Any):
                Add a header to the data. It can be a list or a single string.

        """

        if parse_color:
            try:
                for i in range(len(header)):
                    header[i] = parse_color(header[i])
            except Exception:
                header = header

        self.__headers.append(header)

    def render(self, background_color: Optional[str] = None):
        """
        Render the table. It return a string of table.

        Parameter:
        ---

            background_color (str):
                Set the background color of the table. Color plugin should be installed.

        """

        if parse_color and background_color:
            try:
                table = parse_color(
                    "[bg color={0}]{1}[/bg]".format(
                        background_color,
                        tabulate(
                            self.__rows, headers=self.__headers, tablefmt=self.__type
                        ),
                    )
                )

            except AttributeError:
                raise PluginError(
                    "Table plugin is not supported. Read the docs for more info"
                )

        elif parse_color == None and background_color:
            raise PluginError("Colors are not supported. Read the docs for more info.")

        else:
            try:
                table = tabulate(
                    self.__rows, headers=self.__headers, tablefmt=self.__type
                )
            except AttributeError:
                raise PluginError(
                    "Table plugin is not supported. Read the docs for more info"
                )

        return table

    def __call__(self, background_color: Optional[str] = None):
        """
        Render the table. It return a string of table.

        Parameter:
        ---

            background_color (str):
                Set the background color of the table. Color plugin should be installed.

        """

        return self.render(background_color)



#                   Copyright (c) 2021, Serum Studio
#   Copyright (c) Sindre Sorhus <sindresorhus@gmail.com> (sindresorhus.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#: Spinner frames are https://github.com/sindresorhus

import itertools
import sys
import time
from typing import Optional
from typing import Any
from .constants import SpinnerType
from hype.cursor import hide as hide_cursor
from hype.cursor import show as show_cursor
import threading
from hype.constants import COLOR_SUPPORTED
from hype.constants import rule_colors


class SpinnerNotFound(Exception):
    pass


class SpinnerError(Exception):
    pass


class Spinner:
    """
    A spinner animation based on https://github.com/sindresorhus Spinner Frames.

    Parameters:
    ---

        text (str):
            The text that is printed next to the spinner

        type (str):
            Spinner type. see documentation for more info. or check out `hype.ui.constants`

        cursor (bool):
            Set if the cursor is hidden or shown


    Example:
    ---

        >>> from hype.ui import Spinner
        >>> with Spinner('Loading', type='arc') as spinner:
        >>>     spinner.start() # Start the spinner


    """

    #: Set the stop event to None
    __stop_event = None

    #: Set the stop event to None
    __thread = None

    #: Set the stop event to None
    __thread_id = None

    def __init__(
        self,
        text: Optional[Any] = "",
        type: Optional[str] = "dots",
        cursor: Optional[bool] = False,
        color: Optional[str] = None,
    ):

        self.text = text
        self.type = type
        self.stream = sys.stdout
        self.color = color

        if self.color and COLOR_SUPPORTED == False:
            raise SpinnerError(
                "Colors are not supported. Install using `pip install hypecli[color]`"
            )

        if cursor == False:
            hide_cursor()

    def render(self):
        """
        Render the spinner to the terminal
        """
        if self.type not in SpinnerType.keys():
            raise SpinnerNotFound("%s is not supported." % (self.type))

        frames = SpinnerType[self.type]["frames"]
        interval = 0.001 * SpinnerType[self.type]["interval"]
        spinner = itertools.cycle(frames)

        while not self.__stop_event.set():

            if self.color:
                output = "\r{0}{1}{2} {3}".format(
                    rule_colors[self.color],
                    next(spinner),
                    rule_colors["reset"],
                    self.text,
                )
            else:
                output = "\r{0} {1}".format(next(spinner), self.text)

            self.stream.write(output)
            self.stream.write("\033[K")  #: Clear line
            self.stream.flush()
            self.stream.write("\b")
            time.sleep(interval)

        return self

    @property
    def id(self):
        """
        Get the thread id. For some purposes..
        """
        return self.__thread_id

    @id.setter
    def id(self, value):
        return value

    def start(self):

        self.__thread = threading.Thread(target=self.render)
        self.__stop_event = threading.Event()
        self.__thread_id = self.__thread.name
        self.__thread.setDaemon(True)
        self.__thread.start()

        return self

    def stop(self):
        """Stop the thread on running"""

        if self.__thread and self.__thread.is_alive():
            self.__stop_event.set()

        self.stream.write("\r")
        self.stream.write("\033[K")  #: Clear line
        self.id = None
        show_cursor()
        return self

    def __enter__(self):
        """When the `with` statement opens."""

        return self.start()

    def __exit__(self, type, value, tracerback):
        """When the `with` statement ends"""

        return self.stop()



#                   Copyright (c) 2021, Serum Studio

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import functools
import inspect
import sys

from typing import List, Optional
from typing import Any
from typing import Callable
from typing import Tuple
from typing import get_type_hints

from .command import HypeCommand
from .command import HypeArgument
from .parser import HypeParser

from .print import print as _print

from .constants import rule_colors
from .constants import bg_colors
from .style import Background

from .utils import CommandDict
from .utils import ParamOption
from .utils import OptionDict
from .utils import create_bool_option
from .utils import convert_param_to_option
from .utils import convert_option_to_string

from .errors import ColorNotFound


class Hype:
    """
    The main application for the CLI.

    Parameters:
        -

    Example:

        >>> app = HypeCLI()
        >>> @app.command()
        >>> def greet(name: str):
        >>>     print("hello, {}".format(name))
        >>> ...
        >>> if __name__ == "__main__":
        >>>     app.run()

    """

    __commands = {}
    __required_commands = []
    __commands_function = {}
    __registered_args = {}
    __registered_args_func = {}

    def __init__(self):

        #: The parser object to be used
        self.__parser = HypeParser

    @property
    def commands(self):
        """
        All command registered. Return by it's name.
        """
        command_list = []
        for k in self.__commands.keys():
            command_list.append({self.__commands[k]["name"]: self.__commands[k]['help']})

        return command_list

    def echo(self, text: Any, **options):
        """
        A format wrapper for the hype.print. It can print
        text with colors and styles. using the tag [%tagname%][/]

        Parameters:
        ---
            text (object):
                The value to be printed.

            **options(dict):
                A kwarg of options.

        Available Options:
        ---

        - style
        - foreground
        - color

        Available Color Tags:
        ---

        > For more info check the constant dictionary `hype.constants.rule_colors`

        - blue
        - red
        - yellow
        - magenta
        - green
        - black
        - cyan

        Example:
        ---

            >>> app = Hype()
            >>> app.echo('[red]This is a red color![/]')

        """

        background = options.get("background") or None

        if background in bg_colors:
            _print("%s%s%s" % (bg_colors[background], text, bg_colors["reset"]))

        elif background not in bg_colors and background != None:
            raise ColorNotFound("%s is not yet supported." % (background))

        else:
            _print(text)

    def command(
        self,
        name: str = None,
        usage: Optional[str] = None,
        aliases: Optional[Tuple[Any]] = (),
        help: Optional[str] = ""
    ):

        """
        A command decorator for HypeCommand.

        Parameters:
            name (str):
                The name of the command, if None return the function name

            usage (str):
                The usage format of the command.

            aliases (tuple):
                A tuple of aliases of the command.

            help (str):
                The help format of the command.

        Example:

            >>> @app.command()
            >>> def greet(name: str):
            >>>     \"\"\"
            >>>     Greet the user
            >>>     \"\"\"
            >>>     app.echo(f'Hello {name}')

        """

        #: set the usage of the command
        _usage = usage

        #: Set the aliases of the command
        _aliases = aliases

        def deco(func):
            
            #: Set the name of the command.
            #: If the name is none, return the name of the function

            _name = name if name else func.__name__

            #: Replace the _ to -
            _name = _name.replace('_', '-')

            #: Set the help of the command
            _help = help or func.__doc__

            #: The signature of the function
            signature = inspect.signature(func)

            #: Type hints of the function
            type_hints = func.__annotations__

            #: Set the params to none dict. It should contain the param of the function
            #: and the type hints of the parameters
            params = []
            rargs_keys = {}

            for k,v in self.__registered_args_func.items():
                if k.__name__ == func.__name__:
                    rargs_keys = v
                    
                else:
                    rargs_keys = {}

                    
            for param in signature.parameters.values():
                #: The annotation of the function.
                #: For example: def func(name: str) -> str is the annotaiton
                annotation = param.annotation

                #: The param name of the function
                #: For example: def func(name) -> name is the param.
                param_name = param.name

                #: Replace the _ to - for params
                param_name = param_name.replace('_', '-')


                if param.name in type_hints:
                    annotation = type_hints[param.name]

                if param.name not in rargs_keys:
                    required = True if param.default is inspect.Parameter.empty else False
                    default = (
                        param.default
                        if param.default is not inspect.Parameter.empty
                        else None
                    )
                    anon = annotation if annotation is not inspect.Parameter.empty else None

                    optionparam = ParamOption(
                        convert_param_to_option(param_name),
                        required,
                        default,
                        anon,
                        param.name
                    )
                    params.append(optionparam.to_dict)

            self.__commands[_name] = CommandDict(
                _name, _usage, _help, _aliases, params, func,
            ).to_dict
            self.__commands_function[func] = {"name": _name}

            return func

        return deco

    def argument(self, name: str, type: Optional[Any] = None, help: Optional[str] = "This argument accept anything"):
        """
        A argument decorator for registering arguments. 
        Please take note that when you define argument, make sure
        the name of the argument is on the first parameter of the 
        function.
        
        Parameters:
        ---
            name (str):
                The name of the argument.
            
            type (Optional[Any]):
                The type of the argument.

        Examples:
        ---
            >>> @app.command()
            >>> @app.argument('option')
            >>> def upload(option, debug: bool=False):
            >>>     ...
            >>>     #: Make sure the option argument always on the first one.
            >>>     app.echo(option)

        """
        name = name
        help = help
        type = type 
        
        def deco(func):
            self.__registered_args_func[func] = {name: {'type': type, 'help': help}}
            self.__registered_args[name] = {'type': type, 'help': help}
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return deco

    def exit(self):
        """
        Exit the all application instance without using standard `sys.exit()`

        Parameters:
        ---
            It takes no params yet

        Example:
        ---
            >>> import os
            >>> ...
            >>> @app.command()
            >>> def upload(path: str):
            >>>     if os.path.exists(path):
            >>>         app.echo(f'Path: {path} doesnt exist')
            >>>         app.exit()
            >>>     ...
            >>>     app.echo(f'File Uploaded: {path}')

        """

        try:
            self.__parser.exit()
        except Exception:
            pass

        sys.exit()

    def run(self):
        """
        Run the application.

        Parameters:
        ---
            it takes no params yet

        Example:
        `(More example at the examples folder located on github repo.)`


            >>> from hype import Hype
            >>> app = Hype()
            >>> ...
            >>> @app.command()
            >>> def greet(name: str):
            >>>     app.echo(f'Hello, {name}')
            >>> ...
            >>> @app.command()
            >>> def goodbye(name: str):
            >>>     app.echo(f'Goodbye, {name}')

            >>> if __name__ = "__main__":
            >>>     app.run()

        """

        commands = []
        boolean_options = []
        # self.__registered_args_func = {}
        for k in self.__commands.keys():
            
            for ak, av in self.__registered_args_func.items(): 
                for _k in av.keys():
                    pass

                if self.__commands[k]['func'].__name__ == ak.__name__:                
                    self.__command_parser = HypeCommand(
                        self.__commands[k]["name"],
                        self.__commands[k]["usage"],
                        self.__commands[k]["aliases"],
                        self.__commands[k]["help"],
                        [HypeArgument(name=_k, help=av[_k]['help'], type=av[_k]['type'])]
                    )
                    break

                else:
                    self.__command_parser = HypeCommand(
                        self.__commands[k]["name"],
                        self.__commands[k]["usage"],
                        self.__commands[k]["aliases"],
                        self.__commands[k]["help"]
                    )
                    break
                
            if not self.__registered_args_func:
                self.__command_parser = HypeCommand(
                    self.__commands[k]["name"],
                    self.__commands[k]["usage"],
                    self.__commands[k]["aliases"],
                    self.__commands[k]["help"]
                )
            

            if self.__commands[k]["options"]:
                for _option in self.__commands[k]["options"]:
                    

                    if _option["required"]:

                        self.__required_commands.append(
                            (
                                self.__commands[k]["name"],
                                convert_option_to_string(_option["name"]),
                            )
                        )

                    name = _option["name"]

                    if _option["action"]:

                        bool_name = create_bool_option(_option["name"])
                        for bname in bool_name:

                            if bname == _option["name"]:
                                self.__command_parser.parser.add_option(
                                    _option["name"],
                                    default=_option["default"],
                                    dest=_option["dest"],
                                    action=_option["action"],
                                    metavar=_option["metavar"],
                                )
                            else:
                                self.__command_parser.parser.add_option(
                                    bname,
                                    default=_option["default"],
                                    dest=_option["dest"],
                                    action="store_false",
                                    metavar=_option["metavar"],
                                )

                    else:
                        if isinstance(name, str):

                            self.__command_parser.parser.add_option(
                                name,
                                default=_option["default"],
                                type=_option["type"],
                                dest=_option["dest"],
                                metavar=_option["metavar"],
                            )

                        else:
                            self.__command_parser.parser.add_option(
                                *name,
                                default=_option["default"],
                                type=_option["type"],
                                dest=_option["dest"],
                                metavar=_option["metavar"]
                            )

            commands.append(self.__command_parser)

        parser = self.__parser(commands)

        (
            option,
            command,
            command_opt,
            command_args,
        ) = parser.parse_args()

        params = []

        for i in boolean_options:
            self.__command_parser.parser.add_option(
                i["name"], default=i["default"], type=i["type"], action=i["action"]
            )
            



        if command.name in self.__commands:
            func = self.__commands[command.name]['func']
        

            if command_args:
                # TODO: Check for function registered and return the args
                # NOTE: Please if you have some time improving this, create a pull req 

                for k, v in self.__registered_args_func.items():
                    if k.__name__ == func.__name__:
                        for t in range(len(v)):
                            for _k in v.keys():
                                if v[_k]['type']:
                                    v[_k]['type'](command_args[t])
                            
                            params.append(command_args[t])
                            
            else:
                params.append(None)

            
            for _k, v in vars(command_opt).items():
                if (command.name, _k) in self.__required_commands and v == None:
                    parser.error("Option: {} is required.".format(_k))
                    parser.exit()

                params.append(v)

            func(*params)



class HypeException(Exception):
    def __init__(self, msg=None):
        if msg == None:
            msg = "Something went wrong with the Hype CLI"

        super().__init__(msg)


class OptionError(HypeException):
    def __init__(self, msg="There is something went wrong on the option"):
        super().__init__(msg)


class TooMuchArguments(HypeException):
    def __init__(self, msg="You passed too much arguments"):
        super().__init__(msg)


class PluginError(HypeException):
    def __init__(self, msg="Plugin not installed"):
        super().__init__(msg)


class ColorNotFound(HypeException):
    def __init__(self, msg="The color used is not found"):
        super().__init__(msg)


class TagNotFound(HypeException):
    """
    Used for parsing colors in `hype.colors`
    """

    def __init__(self, msg="Tag is not defined"):
        super().__init__(msg)
